# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import wsme
from wsme import types


class ItemType(types.Base):
    type = wsme.wsattr(types.text, mandatory=True)
    enum = wsme.wsattr([types.text], mandatory=False)

    _wsme_attr_order = ('type', 'enum')

    def __init__(self, **kwargs):
        super(ItemType, self).__init__(**kwargs)
